@@ -1,18 +1,12 @@
 /* We need C99 vsnprintf() semantics */
-#define _ISOC99_SOURCE
-
-/* We need strdup and va_copy */
-
-/* On Linux we have to ask for them
- * On most BSDs they come automatically
- * On other platforms - I have no idea... bug reports welcome :)
- */
-#ifndef _GNU_SOURCE
-# define _GNU_SOURCE
-#endif
-
+#ifdef __GLIBC__
+/* We need C99 vsnprintf() semantics */
+#  define _ISOC99_SOURCE
 /* We need sigaction() and struct sigaction */
-#define _POSIX_C_SOURCE 199309L
+#  define _POSIX_C_SOURCE 199309L
+/* We need strdup and va_copy */
+#  define _GNU_SOURCE
+#endif
 
 #include "tickit.h"
 #include "bindings.h"
@@ -34,6 +28,7 @@
 #include <sys/time.h>
 
 #define streq(a,b) (!strcmp(a,b))
+#define strneq(a,b,n) (strncmp(a,b,n)==0)
 
 /* unit multipliers for working in microseconds */
 #define MSEC      1000
@@ -85,6 +80,8 @@ struct TickitTerm {
 
   int refcount;
   struct TickitBindings bindings;
+
+  int mouse_buttons_held;
 };
 
 DEFINE_BINDINGS_FUNCS(term,TickitTerm,TickitTermEventFn)
@@ -103,13 +100,15 @@ static const char *getstr_hook(const char *name, const char *value, void *_tt)
      *   https://github.com/neovim/neovim/blob/1083c626b9a3fc858c552d38250c3c555cda4074/src/nvim/tui/tui.c#L1982
      */
     struct termios termios;
-    tcgetattr(tt->infd, &termios);
 
-    char *ret = get_tmpbuffer(tt, 2);
-    ret[0] = termios.c_cc[VERASE];
-    ret[1] = 0;
+    if(tt->infd != -1 &&
+       tcgetattr(tt->infd, &termios) == 0) {
+      char *ret = get_tmpbuffer(tt, 2);
+      ret[0] = termios.c_cc[VERASE];
+      ret[1] = 0;
 
-    value = ret;
+      value = ret;
+    }
   }
 
   if(tt->ti_hook.getstr)
@@ -127,7 +126,22 @@ static TermKey *get_termkey(TickitTerm *tt)
     else if(tt->is_utf8 == TICKIT_NO)
       flags |= TERMKEY_FLAG_RAW;
 
-    tt->termkey = termkey_new(tt->infd, TERMKEY_FLAG_EINTR|TERMKEY_FLAG_NOSTART | flags);
+    /* A horrible hack: termkey_new doesn't take a termtype;
+     * termkey_new_abstract does but doesn't take an fd. When we migrate
+     * libtermkey source into here this will be much neater
+     */
+    {
+      const char *was_term = getenv("TERM");
+      setenv("TERM", tt->termtype, true);
+
+      tt->termkey = termkey_new(tt->infd, TERMKEY_FLAG_EINTR|TERMKEY_FLAG_NOSTART | flags);
+
+      if(was_term)
+        setenv("TERM", was_term, true);
+      else
+        unsetenv("TERM");
+    }
+
     termkey_hook_terminfo_getstr(tt->termkey, getstr_hook, tt);
     termkey_start(tt->termkey);
 
@@ -211,6 +225,8 @@ TickitTerm *tickit_term_build(const struct TickitTermBuilder *_builder)
   tt->refcount = 1;
   tt->bindings = (struct TickitBindings){ NULL };
 
+  tt->mouse_buttons_held = 0;
+
   /* Initially empty because we don't necessarily know the initial state
    * of the terminal
    */
@@ -232,6 +248,45 @@ TickitTerm *tickit_term_build(const struct TickitTermBuilder *_builder)
   // Can't 'start' yet until we have an output method
   tt->state = UNSTARTED;
 
+  int fd_in = -1, fd_out = -1;
+
+  switch(builder.open) {
+    case TICKIT_NO_OPEN:
+      break;
+
+    case TICKIT_OPEN_FDS:
+      fd_in  = builder.input_fd;
+      fd_out = builder.output_fd;
+      break;
+
+    case TICKIT_OPEN_STDIO:
+      fd_in  = STDIN_FILENO;
+      fd_out = STDOUT_FILENO;
+      break;
+
+    case TICKIT_OPEN_STDTTY:
+      for(fd_in = 0; fd_in <= 2; fd_in++) {
+        if(isatty(fd_in))
+          break;
+      }
+      if(fd_in > 2) {
+        fprintf(stderr, "Cannot find a TTY filehandle\n");
+        abort();
+      }
+      fd_out = fd_in;
+      break;
+  }
+
+  if(fd_in != -1)
+    tickit_term_set_input_fd(tt, fd_in);
+  if(fd_out != -1)
+    tickit_term_set_output_fd(tt, fd_out);
+  if(builder.output_func)
+    tickit_term_set_output_func(tt, builder.output_func, builder.output_func_user);
+
+  if(builder.output_buffersize)
+    tickit_term_set_output_buffer(tt, builder.output_buffersize);
+
   return tt;
 }
 
@@ -247,21 +302,14 @@ TickitTerm *tickit_term_new_for_termtype(const char *termtype)
   });
 }
 
-TickitTerm *tickit_term_new_for_driver(TickitTermDriver *ttd)
-{
-  return tickit_term_build(&(struct TickitTermBuilder){
-    .driver = ttd,
-  });
-}
-
 TickitTerm *tickit_term_open_stdio(void)
 {
-  TickitTerm *tt = tickit_term_new();
+  TickitTerm *tt = tickit_term_build(&(const struct TickitTermBuilder){
+    .open = TICKIT_OPEN_STDIO,
+  });
   if(!tt)
     return NULL;
 
-  tickit_term_set_input_fd(tt, STDIN_FILENO);
-  tickit_term_set_output_fd(tt, STDOUT_FILENO);
   tickit_term_observe_sigwinch(tt, true);
 
   return tt;
@@ -572,10 +620,6 @@ void tickit_term_await_started_tv(TickitTerm *tt, const struct timeval *timeout)
 
 static void got_key(TickitTerm *tt, TermKey *tk, TermKeyKey *key)
 {
-  if(tt->driver->vtable->gotkey &&
-     (*tt->driver->vtable->gotkey)(tt->driver, tk, key))
-    return;
-
   if(key->type == TERMKEY_TYPE_MOUSE) {
     TermKeyMouseEvent ev;
     TickitMouseEventInfo info;
@@ -597,6 +641,23 @@ static void got_key(TickitTerm *tt, TermKey *tk, TermKeyKey *key)
 
     info.mod = key->modifiers;
 
+    if(info.type == TICKIT_MOUSEEV_PRESS || info.type == TICKIT_MOUSEEV_DRAG) {
+      tt->mouse_buttons_held |= (1 << info.button);
+    }
+    else if(info.type == TICKIT_MOUSEEV_RELEASE && info.button) {
+      tt->mouse_buttons_held &= ~(1 << info.button);
+    }
+    else if(info.type == TICKIT_MOUSEEV_RELEASE) {
+      /* X10 cannot report which button was released. Just report that they 
+       * all were */
+      for(info.button = 1; tt->mouse_buttons_held; info.button++)
+        if(tt->mouse_buttons_held & (1 << info.button)) {
+          run_events_whilefalse(tt, TICKIT_TERM_ON_MOUSE, &info);
+          tt->mouse_buttons_held &= ~(1 << info.button);
+        }
+      return; // Buttons have been handled
+    }
+
     run_events_whilefalse(tt, TICKIT_TERM_ON_MOUSE, &info);
   }
   else if(key->type == TERMKEY_TYPE_UNICODE && !key->modifiers) {
@@ -623,6 +684,24 @@ static void got_key(TickitTerm *tt, TermKey *tk, TermKeyKey *key)
 
     run_events_whilefalse(tt, TICKIT_TERM_ON_KEY, &info);
   }
+  else if(key->type == TERMKEY_TYPE_MODEREPORT) {
+    if(tt->driver->vtable->on_modereport) {
+      int initial, mode, value;
+      termkey_interpret_modereport(tk, key, &initial, &mode, &value);
+
+      (tt->driver->vtable->on_modereport)(tt->driver, initial, mode, value);
+    }
+  }
+  else if(key->type == TERMKEY_TYPE_DCS) {
+    const char *dcs;
+    if(termkey_interpret_string(tk, key, &dcs) != TERMKEY_RES_KEY)
+      return;
+
+    if(strneq(dcs, "1$r", 3)) { // Successful DECRQSS
+      if(tt->driver->vtable->on_decrqss)
+        (tt->driver->vtable->on_decrqss)(tt->driver, dcs + 3, strlen(dcs + 3));
+    }
+  }
 }
 
 void tickit_term_emit_key(TickitTerm *tt, TickitKeyEventInfo *info)
